@@ -10,10 +10,10 @@ use Zonemaster::Engine;
 use Zonemaster::Engine::Packet;
 use Zonemaster::Engine::Nameserver::Cache;
 use Zonemaster::Engine::Recursor;
-use Zonemaster::Engine::Constants ':misc';
+use Zonemaster::Engine::Constants qw( :ip :misc );
 use Zonemaster::LDNS;
 
-use Zonemaster::Engine::Net::IP;
+use Net::IP::XS;
 use Time::HiRes qw[time];
 use JSON::PP;
 use MIME::Base64;
@@ -37,7 +37,9 @@ has 'dns'   => ( is => 'ro' );
 has 'cache' => ( is => 'ro' );
 has 'times' => ( is => 'ro' );
 
-has 'source_address' => ( is => 'ro' );
+has 'source_address'  => ( is => 'ro' );
+has 'source_address4' => ( is => 'ro' );
+has 'source_address6' => ( is => 'ro' );
 
 has 'fake_delegations' => ( is => 'ro' );
 has 'fake_ds'          => ( is => 'ro' );
@@ -62,13 +64,15 @@ sub new {
     my $attrs = shift;
 
     my %lazy_attrs;
-    $lazy_attrs{source_address} = delete $attrs->{source_address} if exists $attrs->{source_address};
-    $lazy_attrs{dns}            = delete $attrs->{dns}            if exists $attrs->{dns};
-    $lazy_attrs{cache}          = delete $attrs->{cache}          if exists $attrs->{cache};
+    $lazy_attrs{source_address}  = delete $attrs->{source_address}  if exists $attrs->{source_address};
+    $lazy_attrs{source_address4} = delete $attrs->{source_address4} if exists $attrs->{source_address4};
+    $lazy_attrs{source_address6} = delete $attrs->{source_address6} if exists $attrs->{source_address6};
+    $lazy_attrs{dns}             = delete $attrs->{dns}             if exists $attrs->{dns};
+    $lazy_attrs{cache}           = delete $attrs->{cache}           if exists $attrs->{cache};
 
     # Required arguments
     confess "Attribute \(address\) is required"
-      if !exists $attrs->{address};
+      if !defined $attrs->{address};
 
     # Type coercions
     $attrs->{name} = Zonemaster::Engine::DNSName->from_string( $attrs->{name} )
@@ -80,9 +84,9 @@ sub new {
     my $address;
 
     # Use a object cache for IP type coercion (don't parse IP unless it is needed)
-    if (!blessed $attrs->{address} || !$attrs->{address}->isa( 'Zonemaster::Engine::Net::IP' )) {
+    if (!blessed $attrs->{address} || !$attrs->{address}->isa( 'Net::IP::XS' )) {
         if (!exists $address_object_cache{$attrs->{address}}) {
-            $address_object_cache{$attrs->{address}} = Zonemaster::Engine::Net::IP->new($attrs->{address});
+            $address_object_cache{$attrs->{address}} = Net::IP::XS->new($attrs->{address});
             $address_repr_cache{$attrs->{address}} = $address_object_cache{$attrs->{address}}->ip;
         }
         # Fetch IP object from the address cache (avoid object creation and method call)
@@ -100,9 +104,9 @@ sub new {
     # Type constraints
     confess "Argument must be coercible into a Zonemaster::Engine::DNSName: name"
       if !$attrs->{name}->isa( 'Zonemaster::Engine::DNSName' );
-    confess "Argument must be coercible into a Zonemaster::Engine::Net::IP: address"
+    confess "Argument must be coercible into a Net::IP::XS: address"
       if exists $attrs->{address}
-      && !$attrs->{address}->isa( 'Zonemaster::Engine::Net::IP' );
+      && !$attrs->{address}->isa( 'Net::IP::XS' );
     confess "Argument must be an ARRAYREF: times"
       if exists $attrs->{times}
       && ref $attrs->{times} ne 'ARRAY';
@@ -118,6 +122,12 @@ sub new {
     confess "Argument must be a string or undef: source_address"
       if exists $lazy_attrs{source_address}
       && ref $lazy_attrs{source_address} ne '';
+    confess "Argument must be a string or undef: source_address4"
+      if exists $lazy_attrs{source_address4}
+      && ref $lazy_attrs{source_address4} ne '';
+    confess "Argument must be a string or undef: source_address6"
+      if exists $lazy_attrs{source_address6}
+      && ref $lazy_attrs{source_address6} ne '';
     confess "Argument must be a Zonemaster::LDNS: dns"
       if exists $lazy_attrs{dns}
       && ( !blessed $lazy_attrs{dns} || !$lazy_attrs{dns}->isa( 'Zonemaster::LDNS' ) );
@@ -132,9 +142,11 @@ sub new {
     $attrs->{times}            //= [];
 
     my $obj = Class::Accessor::new( $class, $attrs );
-    $obj->{_source_address} = $lazy_attrs{source_address} if exists $lazy_attrs{source_address};
-    $obj->{_dns}            = $lazy_attrs{dns}            if exists $lazy_attrs{dns};
-    $obj->{_cache}          = $lazy_attrs{cache}          if exists $lazy_attrs{cache};
+    $obj->{_source_address}  = $lazy_attrs{source_address}  if exists $lazy_attrs{source_address};
+    $obj->{_source_address4} = $lazy_attrs{source_address4} if exists $lazy_attrs{source_address4};
+    $obj->{_source_address6} = $lazy_attrs{source_address6} if exists $lazy_attrs{source_address6};
+    $obj->{_dns}             = $lazy_attrs{dns}             if exists $lazy_attrs{dns};
+    $obj->{_cache}           = $lazy_attrs{cache}           if exists $lazy_attrs{cache};
 
     Zonemaster::Engine->logger->add( NS_CREATED => { name => $name, ip => $obj->address->ip } );
     $object_cache{$name}{$address} = $obj;
@@ -143,9 +155,9 @@ sub new {
 }
 
 sub source_address {
-    my $self = shift;
+    my ( $self, $ip_version ) = @_;
 
-    # Lazy default value
+    # Lazy default values
     if ( !exists $self->{_source_address} ) {
         my $value = Zonemaster::Engine::Profile->effective->get( q{resolver.source} );
         if ( $value eq $RESOLVER_SOURCE_OS_DEFAULT ) {
@@ -155,6 +167,21 @@ sub source_address {
             $self->{_source_address} = $value;
         }
     }
+    if ( !exists $self->{_source_address4} ) {
+        my $value = Zonemaster::Engine::Profile->effective->get( q{resolver.source4} );
+        $self->{_source_address4} = $value eq '' ? undef : $value;
+    }
+    if ( !exists $self->{_source_address6} ) {
+        my $value = Zonemaster::Engine::Profile->effective->get( q{resolver.source6} );
+        $self->{_source_address6} = $value eq '' ? undef : $value;
+    }
+
+    if ( $ip_version == $IP_VERSION_4 and $self->{_source_address4} ) {
+        return $self->{_source_address4};
+    }
+    if ( $ip_version == $IP_VERSION_6 and $self->{_source_address6} ) {
+        return $self->{_source_address6};
+    }
 
     return $self->{_source_address};
 }
@@ -185,20 +212,23 @@ sub _build_dns {
     my ( $self ) = @_;
 
     my $res = Zonemaster::LDNS->new( $self->address->ip );
+    
     $res->recurse( 0 );
+    $res->dnssec( 0 );
+    $res->edns_size( $UDP_EDNS_QUERY_DEFAULT );
 
     $res->retry( Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.retry} ) );
     $res->retrans( Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.retrans} ) );
-    $res->dnssec( Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.dnssec} ) );
     $res->usevc( Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.usevc} ) );
     $res->igntc( Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.igntc} ) );
     $res->recurse( Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.recurse} ) );
     $res->debug( Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.debug} ) );
-    $res->edns_size( Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.edns_size} ) );
     $res->timeout( Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.timeout} ) );
 
-    if ( $self->source_address ) {
-        $res->source( $self->source_address );
+    my $ip_version = Net::IP::XS::ip_get_version( $self->address->ip );
+    my $source_address = $self->source_address( $ip_version );
+    if ( $source_address ) {
+        $res->source( $source_address );
     }
 
     return $res;
@@ -207,7 +237,10 @@ sub _build_dns {
 sub _build_cache {
     my ( $self ) = @_;
 
-    Zonemaster::Engine::Nameserver::Cache->new( { address => $self->address } );
+    my $cache_type = Zonemaster::Engine::Nameserver::Cache->get_cache_type( Zonemaster::Engine::Profile->effective );
+    my $cache_class = Zonemaster::Engine::Nameserver::Cache->get_cache_class( $cache_type );
+
+    $cache_class->new( { address => $self->address } );
 }
 
 ###
@@ -229,7 +262,7 @@ sub query {
     }
 
     Zonemaster::Engine->logger->add(
-        'query',
+        'QUERY',
         {
             name  => "$name",
             type  => $type,
@@ -239,22 +272,27 @@ sub query {
     );
 
     my $class     = $href->{class}     // 'IN';
-    my $dnssec    = $href->{dnssec}    // Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.dnssec} );
+    my $dnssec    = $href->{dnssec}    // 0;
     my $usevc     = $href->{usevc}     // Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.usevc} );
     my $recurse   = $href->{recurse}   // Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.recurse} );
-    my $edns_size = $href->{edns_size} // Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.edns_size} );
+    my $edns_size = $href->{edns_size} // $UDP_EDNS_QUERY_DEFAULT;
 
     # Fake a DS answer
     if ( $type eq 'DS' and $class eq 'IN' and $self->fake_ds->{ lc( $name ) } ) {
         my $p = Zonemaster::LDNS::Packet->new( $name, $type, $class );
+
+        $p->qr( 1 );
         $p->aa( 1 );
         $p->do( $dnssec );
         $p->rd( $recurse );
+
         foreach my $rr ( @{ $self->fake_ds->{ lc( $name ) } } ) {
             $p->unique_push( 'answer', $rr );
         }
+
         my $res = Zonemaster::Engine::Packet->new( { packet => $p } );
-        Zonemaster::Engine->logger->add( FAKE_DS_RETURNED => { name => "$name", from => "$self" } );
+        Zonemaster::Engine->logger->add( FAKE_DS_RETURNED => { name => "$name", type  => $type, class => $class, from => "$self" } );
+        Zonemaster::Engine->logger->add( FAKE_PACKET_RETURNED => { packet => $res->string } );
         return $res;
     }
 
@@ -269,74 +307,64 @@ sub query {
                 $p->unique_push( 'answer',     $_ ) for @{$name};
                 $p->unique_push( 'additional', $_ ) for @{$addr};
             }
+            elsif ( $type eq 'DS' ) {
+                $p->aa( 1 );
+            }
             else {
                 while ( my ( $section, $aref ) = each %{ $self->fake_delegations->{$fname} } ) {
                     $p->unique_push( $section, $_ ) for @{$aref};
                 }
             }
 
-            $p->aa( 0 );
+            $p->aa( 0 ) unless ( $type eq 'DS' );
+            $p->qr( 1 );
             $p->do( $dnssec );
             $p->rd( $recurse );
             $p->answerfrom( $self->address->ip );
-            Zonemaster::Engine->logger->add(
-                'FAKE_DELEGATION',
-                {
-                    name  => "$name",
-                    type  => $type,
-                    class => $class,
-                    from  => "$self",
-                }
-            );
+
+            Zonemaster::Engine->logger->add( FAKE_DELEGATION_RETURNED => { name  => "$name", type  => $type, class => $class, from  => "$self" } );
 
             my $res = Zonemaster::Engine::Packet->new( { packet => $p } );
-            Zonemaster::Engine->logger->add( FAKED_RETURN => { packet => $res->string } );
+            Zonemaster::Engine->logger->add( FAKE_PACKET_RETURNED => { packet => $res->string } );
             return $res;
         } ## end if ( $name =~ m/([.]|\A)\Q$fname\E\z/xi)
     } ## end foreach my $fname ( sort keys...)
 
     my $p;
     my $md5 = Digest::MD5->new;
-    my $edns_special_case = 0;
-    if ( defined $href->{edns_details} ) {
-        if ( defined $href->{edns_details}{version} and $href->{edns_details}{version} != 0 ) {
-            $edns_special_case = 1;
-        }
-        elsif ( defined $href->{edns_details}{z} ) {
-            $edns_special_case = 1;
-        }
-        elsif ( defined $href->{edns_details}{extended_rcode} ) {
-            $edns_special_case = 1;
-        }
-        elsif ( defined $href->{edns_details}{data} ) {
-            $edns_special_case = 1;
-        }
-        elsif ( defined $href->{edns_details}{udp_size} ) {
-            $edns_size = $href->{edns_details}{udp_size};
-        }
-    }
 
     $md5->add( q{NAME}    , $name );
     $md5->add( q{TYPE}    , "\U$type" );
     $md5->add( q{CLASS}   , "\U$class" );
-    $md5->add( q{DNSSEC}  , $dnssec );
-    $md5->add( q{USEVC}   , $usevc );
-    $md5->add( q{RECURSE} , $recurse );
-    if ( $edns_special_case ) {
-        $md5->add( q{EDNS_VERSION}        , $href->{edns_details}{version} ? $href->{edns_details}{version} : 0 );
-        $md5->add( q{EDNS_Z}              , $href->{edns_details}{z} ? $href->{edns_details}{z} : 0 );
-        $md5->add( q{EDNS_EXTENDED_RCODE} , $href->{edns_details}{extended_rcode} ? $href->{edns_details}{extended_rcode} : 0 );
-        $md5->add( q{EDNS_DATA}           , $href->{edns_details}{data} ? $href->{edns_details}{data} : q{} );
-        $md5->add( q{EDNS_UDP_SIZE}       , $href->{edns_details}{udp_size} ? $href->{edns_details}{udp_size} : 0 );
+
+    if ( exists $href->{edns_details} and exists $href->{edns_details}{do} ) {
+        $md5->add( q{DNSSEC} , $href->{edns_details}{do} );
     }
     else {
-        $md5->add( q{EDNS_UDP_SIZE}       , $edns_size);
+        $md5->add( q{DNSSEC} , $dnssec );
     }
+
+    $md5->add( q{USEVC}   , $usevc );
+    $md5->add( q{RECURSE} , $recurse );
+
+    if ( exists $href->{edns_details} ) {
+        $md5->add( q{EDNS_VERSION}        , $href->{edns_details}{version} // 0 );
+        $md5->add( q{EDNS_Z}              , $href->{edns_details}{z} // 0 );
+        $md5->add( q{EDNS_EXTENDED_RCODE} , $href->{edns_details}{rcode} // 0 );
+        $md5->add( q{EDNS_DATA}           , $href->{edns_details}{data} // q{} );
+        $md5->add( q{EDNS_UDP_SIZE}       , $href->{edns_details}{size} // $edns_size );
+    }
+    else {
+        $md5->add( q{EDNS_UDP_SIZE} , 0);
+    }
+
     my $idx = $md5->b64digest();
-    if ( not exists( $self->cache->data->{$idx} ) ) {
-        $self->cache->data->{$idx} = $self->_query( $name, $type, $href, $edns_special_case );
+
+    my ( $in_cache, $p) = $self->cache->get_key( $idx );
+    if ( not $in_cache ) {
+        $p = $self->_query( $name, $type, $href );
+        $self->cache->set_key( $idx, $p );
     }
-    $p = $self->cache->data->{$idx};
 
     Zonemaster::Engine->logger->add( CACHED_RETURN => { packet => ( $p ? $p->string : 'undef' ) } );
 
@@ -351,19 +379,18 @@ sub add_fake_delegation {
     foreach my $name ( keys %{$href} ) {
         push @{ $delegation{authority} }, Zonemaster::LDNS::RR->new( sprintf( '%s IN NS %s', $domain, $name ) );
         foreach my $ip ( @{ $href->{$name} } ) {
-            if ( Zonemaster::Engine::Net::IP->new( $ip )->ip eq $self->address->ip ) {
-                Zonemaster::Engine->logger->add(
-                    FAKE_DELEGATION_TO_SELF => { ns => "$self", domain => $domain, data => $href } );
+            if ( Net::IP::XS->new( $ip )->ip eq $self->address->ip ) {
+                Zonemaster::Engine->logger->add( FAKE_DELEGATION_TO_SELF => { ns => "$self", domain => $domain, data => $href } );
                 return;
             }
 
             push @{ $delegation{additional} },
-              Zonemaster::LDNS::RR->new( sprintf( '%s IN %s %s', $name, ( Zonemaster::Engine::Net::IP::ip_is_ipv6( $ip ) ? 'AAAA' : 'A' ), $ip ) );
+              Zonemaster::LDNS::RR->new( sprintf( '%s IN %s %s', $name, ( Net::IP::XS::ip_is_ipv6( $ip ) ? 'AAAA' : 'A' ), $ip ) );
         }
     }
 
     $self->fake_delegations->{$domain} = \%delegation;
-    Zonemaster::Engine->logger->add( ADDED_FAKE_DELEGATION => { ns => "$self", domain => $domain, data => $href } );
+    Zonemaster::Engine->logger->add( FAKE_DELEGATION_ADDED => { ns => "$self", domain => $domain, data => $href } );
 
     # We're changing the world, so the cache can't be trusted
     Zonemaster::Engine::Recursor->clear_cache;
@@ -379,7 +406,6 @@ sub add_fake_ds {
         $domain = Zonemaster::Engine::DNSName->new( $domain );
     }
 
-    Zonemaster::Engine->logger->add( FAKE_DS => { domain => lc( "$domain" ), data => $aref, ns => "$self" } );
     foreach my $href ( @{$aref} ) {
         push @ds,
           Zonemaster::LDNS::RR->new(
@@ -391,6 +417,7 @@ sub add_fake_ds {
     }
 
     $self->fake_ds->{ lc( "$domain" ) } = \@ds;
+    Zonemaster::Engine->logger->add( FAKE_DS_ADDED => { domain => lc( "$domain" ), data => $aref, ns => "$self" } );
 
     # We're changing the world, so the cache can't be trusted
     Zonemaster::Engine::Recursor->clear_cache;
@@ -399,7 +426,7 @@ sub add_fake_ds {
 } ## end sub add_fake_ds
 
 sub _query {
-    my ( $self, $name, $type, $href, $edns_special_case ) = @_;
+    my ( $self, $name, $type, $href ) = @_;
     my %flags;
 
     $type //= 'A';
@@ -424,16 +451,19 @@ sub _query {
     # Make sure we have a value for each flag
     $flags{q{retry}}     = $href->{q{retry}}     // Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.retry} );
     $flags{q{retrans}}   = $href->{q{retrans}}   // Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.retrans} );
-    $flags{q{dnssec}}    = $href->{q{dnssec}}    // Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.dnssec} );
+    $flags{q{dnssec}}    = $href->{q{dnssec}}    // 0;
     $flags{q{usevc}}     = $href->{q{usevc}}     // Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.usevc} );
     $flags{q{igntc}}     = $href->{q{igntc}}     // Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.igntc} );
     $flags{q{fallback}}  = $href->{q{fallback}}  // Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.fallback} );
     $flags{q{recurse}}   = $href->{q{recurse}}   // Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.recurse} );
-    $flags{q{edns_size}} = $href->{q{edns_size}} // Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.edns_size} );
     $flags{q{timeout}}   = $href->{q{timeout}}   // Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.timeout} );
-    if ( defined $href->{edns_details} and $href->{edns_details}{udp_size} ) {
-        $flags{q{edns_size}} = $href->{edns_details}{udp_size};
+    $flags{q{edns_size}} = $href->{q{edns_size}} // $UDP_EDNS_QUERY_DEFAULT;
+
+    if ( exists $href->{edns_details} ) {
+        $flags{q{dnssec}}    = $href->{edns_details}{do} // $flags{q{dnssec}};
+        $flags{q{edns_size}} = $href->{edns_details}{size} // $flags{q{edns_size}};
     }
+
     # Set flags for this query
     foreach my $flag ( keys %flags ) {
         $self->dns->$flag( $flags{$flag} );
@@ -455,37 +485,35 @@ sub _query {
         );
     }
     else {
-        if ( $edns_special_case ) {
+        if ( exists $href->{edns_details} ) {
             my $pkt = Zonemaster::LDNS::Packet->new("$name", $type, $href->{class} );
-            if ( defined $href->{edns_details} and defined $href->{edns_details}{version} and $href->{edns_details}{version} != 0 ) {
-                $pkt->set_edns_present();
+            $pkt->set_edns_present();
+
+            if ( exists $href->{edns_details}{version} ) {
                 $pkt->edns_version($href->{edns_details}{version});
             }
-	    if ( defined $href->{edns_details} and defined $href->{edns_details}{z} ) {
-                $pkt->set_edns_present();
+    	    if ( exists $href->{edns_details}{z} ) {
                 $pkt->edns_z($href->{edns_details}{z});
             }
-	    if ( defined $href->{edns_details} and defined $href->{edns_details}{do} ) {
-                $pkt->set_edns_present();
+    	    if ( exists $href->{edns_details}{do} ) {
                 $pkt->do($href->{edns_details}{do});
             }
-	    if ( defined $href->{edns_details} and defined $href->{edns_details}{udp_size} ) {
-                $pkt->set_edns_present();
-                $pkt->edns_size($href->{edns_details}{udp_size});
+    	    if ( exists $href->{edns_details}{size} ) {
+                $pkt->edns_size($href->{edns_details}{size});
             }
-	    if ( defined $href->{edns_details} and defined $href->{edns_details}{extended_rcode} ) {
-                $pkt->set_edns_present();
-                $pkt->edns_rcode($href->{edns_details}{extended_rcode});
+    	    if ( exists $href->{edns_details}{rcode} ) {
+                $pkt->edns_rcode($href->{edns_details}{rcode});
             }
-            if ( defined $href->{edns_details} and defined $href->{edns_details}{data} ) {
-                $pkt->set_edns_present();
+            if ( exists $href->{edns_details}{data} ) {
                 $pkt->edns_data($href->{edns_details}{data});
             }
-	    $res = eval { $self->dns->query_with_pkt( $pkt ) };
+
+    	    $res = eval { $self->dns->query_with_pkt( $pkt ) };
         }
         else {
             $res = eval { $self->dns->query( "$name", $type, $href->{class} ) };
         }
+
         if ( $@ ) {
             my $msg = "$@";
             my $trailing_info = " at ".__FILE__;
@@ -504,9 +532,11 @@ sub _query {
     push @{ $self->times }, ( time() - $before );
 
     # Reset to defaults
-
     foreach my $flag ( keys %flags ) {
-        $self->dns->$flag( Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.}.$flag ) );
+        # Except for any flag that is not configurable in the profile
+        unless ( grep( /^$flag$/, ( 'dnssec', 'edns_size' ) ) ) {
+            $self->dns->$flag( Zonemaster::Engine::Profile->effective->get( q{resolver.defaults.}.$flag ) );
+        }
     }
 
     if ( $res ) {
@@ -581,6 +611,9 @@ sub restore {
         }
       );
 
+    my $cache_type = Zonemaster::Engine::Nameserver::Cache->get_cache_type( Zonemaster::Engine::Profile->effective );
+    my $cache_class = Zonemaster::Engine::Nameserver::Cache->get_cache_class( $cache_type );
+
     open my $fh, '<', $filename or die "Failed to open restore data file: $!\n";
     while ( my $line = <$fh> ) {
         my ( $name, $addr, $data ) = split( / /, $line, 3 );
@@ -588,8 +621,8 @@ sub restore {
         my $ns  = Zonemaster::Engine::Nameserver->new(
             {
                 name    => $name,
-                address => Zonemaster::Engine::Net::IP->new($addr),
-                cache   => Zonemaster::Engine::Nameserver::Cache->new( { data => $ref, address => Zonemaster::Engine::Net::IP->new( $addr ) } )
+                address => Net::IP::XS->new($addr),
+                cache   => $cache_class->new( { data => $ref, address => Net::IP::XS->new( $addr ) } )
             }
         );
     }
@@ -730,7 +763,7 @@ A L<Zonemaster::Engine::DNSName> object holding the nameserver's name.
 
 =item address
 
-A L<Zonemaster::Engine::Net::IP> object holding the nameserver's address.
+A L<Net::IP::XS> object holding the nameserver's address.
 
 =item dns
 
@@ -743,6 +776,17 @@ A reference to a L<Zonemaster::Engine::Nameserver::Cache> object holding the cac
 =item source_address
 
 The source address all resolver objects should use when sending queries.
+Depends on the IP version used to send the queries.
+
+=item source_address4
+
+The IPv4 source address all resolver objects should use when sending queries
+over IPv4.
+
+=item source_address6
+
+The IPv6 source address all resolver objects should use when sending queries
+over IPv6.
 
 =item times
 
@@ -785,7 +829,7 @@ Remove all cached nameserver objects and queries.
 
 Send a DNS query to the nameserver the object represents. C<$name> and C<$type> are the name and type that will be queried for (C<$type> defaults
 to 'A' if it's left undefined). C<$flagref> is a reference to a hash, the keys of which are flags and the values are their corresponding values.
-The available flags are as follows. All but the first directly correspond to methods in the L<Zonemaster::LDNS::Resolver> object.
+The available flags are as follows. All but 'class' and 'edns_details' directly correspond to methods in the L<Zonemaster::LDNS::Resolver> object.
 
 =over
 
@@ -799,11 +843,14 @@ Send the query via TCP (only).
 
 =item retrans
 
-The retransmission interval
+The retransmission interval.
 
 =item dnssec
 
-Set the DO flag in the query.
+Set the DO flag in the query. Defaults to false.
+
+If set to true, it becomes an EDNS query.
+Value overridden by 'edns_details->do' (if also given). More details in 'edns_details' below.
 
 =item debug
 
@@ -833,6 +880,26 @@ If set to true, incoming response packets with the TC flag set fall back to EDNS
 
 If set to true, prevents a server to be black-listed on a query in case there is no answer OR rcode is REFUSED.
 
+=item edns_size
+
+Set the EDNS0 UDP maximum size. Defaults to 512.
+
+Used only when the query is an EDNS query. Does not enable on its own the query to be an EDNS query.
+Value overridden by 'edns_details->size' (if also given). More details in 'edns_details' below.
+
+=item edns_details
+
+A hash. An empty hash or a hash with any keys below will enable the query to be an EDNS query.
+
+The currently supported keys are 'version', 'z', 'do', 'rcode', 'size' and 'data'.
+See L<Zonemaster::LDNS::Packet> for more details (key names prefixed with 'edns_').
+
+Note that flag 'edns_size' also exists (see above) and has the same effect as 'edns_details->size', although the value of the 
+latter will take precedence if both are given.
+
+Similarly, note that flag 'dnssec' also exists (see above) and has the same effect as 'edns_details->do', although the value of the
+latter will take precedence if both are given.
+
 =back
 
 =item string()
